/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __geometry_MatrixSqrt_h__
#define __geometry_MatrixSqrt_h__

#define	FE_MSQ_DEBUG		FALSE
#define	FE_MSQ_VERIFY		(FE_CODEGEN<=FE_DEBUG)
#define	FE_MSQ_MAX_ERROR	1e-3

namespace fe
{

/**************************************************************************//**
	@brief solve A=B*B for B, given A

	@ingroup geometry

	Uses Denman-Beavers square root iteration.
*//***************************************************************************/
template <typename MATRIX>
class MatrixSqrt
{
	public:
				MatrixSqrt(void):
					m_iterations(8)											{}

		void	solve(MATRIX& B, const MATRIX& A) const;

		void	setIterations(U32 iterations)	{ m_iterations=iterations; }

	private:
		U32		m_iterations;
};

template <typename MATRIX>
inline void MatrixSqrt<MATRIX>::solve(MATRIX& B, const MATRIX& A) const
{
#if FE_MSQ_DEBUG
	feLog("\nA\n%s\n",print(A).c_str());
#endif

	MATRIX Y[2];
	MATRIX Z[2];
	MATRIX invY;
	MATRIX invZ;
	U32 current=0;

	Y[0]=A;
	setIdentity(Z[0]);

	U32 iteration;
	for(iteration=0;iteration<m_iterations;iteration++)
	{
		U32 last=current;
		current= !current;

#if FE_MSQ_DEBUG
		feLog("iteration %d\nY\n%s\nZ\n%s\n",
				iteration,print(Y[last]).c_str(),print(Z[last]).c_str());
#endif

		invert(invY,Y[last]);
		invert(invZ,Z[last]);

#if FE_MSQ_DEBUG
		feLog("invY\n%s\ninvZ\n%s\n",
				print(invY).c_str(),print(invZ).c_str());
#endif

		Y[current]=0.5*(Y[last]+invZ);
		Z[current]=0.5*(Z[last]+invY);
	}

	MATRIX& R=Y[current];

#if FE_MSQ_DEBUG
	feLog("\nA\n%s\n",print(A).c_str());
	feLog("\nB\n%s\nB*B\n%s\n",print(R).c_str(),print(R*R).c_str());
#endif

#if FE_MSQ_VERIFY
	BWORD invalid=FALSE;
	MATRIX diff=A-R*R;
	F32 sum=0.0f;
	for(U32 m=0;m<width(diff);m++)
	{
		for(U32 n=0;n<height(diff);n++)
		{
			if(FE_INVALID_SCALAR(diff(m,n)))
			{
				invalid=TRUE;
			}
			sum+=fabs(diff(m,n));
		}
	}
#endif
#if FE_MSQ_VERIFY && FE_MSQ_DEBUG
	feLog("\ndiff\n%s\ncomponent sum=%.6G\n",print(diff).c_str(),sum);
#endif
#if FE_MSQ_VERIFY
	if(invalid || sum>FE_MSQ_MAX_ERROR)
	{
		feLog("MatrixSqrt<>::solve error of %.6G exceeded limit of %.6G\n",
				sum, FE_MSQ_MAX_ERROR);
		feLog("\nA\n%s\n",print(A).c_str());
		feLog("\nB\n%s\nB*B\n%s\n",print(R).c_str(),print(R*R).c_str());

		feX("MatrixSqrt<>::solve","failed to converge");
	}
#endif

	B=R;
}

} /* namespace */

#endif /* __geometry_MatrixSqrt_h__ */
