/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "operator/operator.pmh"

using namespace fe;

void SurfaceBindOp::initialize(void)
{
	catalog<bool>("Bind Primitives")=true;
	catalog<String>("Bind Primitives","hint")=
			"Affect whole primitives instead of individual points";

	catalog<bool>("Spider")=false;
	catalog<String>("Spider","label")="Spider Wrap";
	catalog<String>("Spider","hint")=
			"Track a triangle of distant points.";

	catalog<I32>("Bindings")=1;
	catalog<I32>("Bindings","min")=1;
	catalog<String>("Bindings","hint")=
			"Maximum number of attachment contacts";

	catalog<String>("Bind Attribute Prefix")="bind";
	catalog<String>("Bind Attribute Prefix","hint")=
			"Prefix of attribute names added to output surface describing"
			" where each point on the input surface is attached to"
			" the associated driver";

	catalog< sp<Component> >("Input Surface");
	catalog< sp<Component> >("Driver Surface");
}

void SurfaceBindOp::handle(Record& a_rSignal)
{
//	feLog("SurfaceBindOp::handle\n");

	//* access output surface (copy of input surface)
	sp<SurfaceAccessibleI> spOutputAccessible=surfaceOutput(a_rSignal);
	if(spOutputAccessible.isNull())
	{
		feLog("SurfaceBindOp::handle"
				" output surface not SurfaceAccessibleI\n");
		catalog<String>("error")="inaccessible output";
		return;
	}

	//* access reference curves
	sp<SurfaceAccessibleI> spDriverAccessible=
			catalog< sp<Component> >("Driver Surface");
	if(spDriverAccessible.isNull())
	{
		feLog("SurfaceBindOp::handle"
				" driver surface not SurfaceAccessibleI\n");
		catalog<String>("error")="inaccessible driver surface";
		return;
	}
	sp<SurfaceI> spDriver=spDriverAccessible->surface();
	if(spDriver.isNull())
	{
		feLog("SurfaceBindOp::handle"
				" driver surface failed to provide SurfaceI\n");
		catalog<String>("error")="invalid driver surface";
		return;
	}
	spDriver->setRefinement(0);

	const bool bindPrimitives=catalog<bool>("Bind Primitives");
	const bool spiderWrap=catalog<bool>("Spider");

	sp<SurfaceAccessorI> spOutputPrimitives;
	if(bindPrimitives || spiderWrap)
	{
		spOutputPrimitives=spOutputAccessible->accessor(
				SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
		FEASSERT(spOutputPrimitives.isValid());
	}

	sp<SurfaceAccessorI> spOutputElements;
	if(bindPrimitives)
	{
		spOutputElements=spOutputPrimitives;
	}
	else
	{
		spOutputElements=spOutputAccessible->accessor(
				SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position);
		FEASSERT(spOutputElements.isValid());
	}

	sp<SurfaceAccessorI> spOutputNormals=spOutputAccessible->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_normal);
	FEASSERT(spOutputNormals.isValid());

	std::vector< sp<SurfaceAccessorI> > outputFaces;
	std::vector< sp<SurfaceAccessorI> > outputBarys;
	const I32 bindings=accessBindings(spOutputAccessible,
			outputFaces,outputBarys,TRUE);
	if(bindings<1)
	{
		return;
	}

	I32 face= -1;
	Barycenter<Real> barycenter;

	const U32 elementCount=spOutputElements->count();

	SpatialVector* pOneEdge=NULL;
	if(spiderWrap)
	{
		pOneEdge=new SpatialVector[elementCount];
		for(U32 index=0;index<elementCount;index++)
		{
			set(pOneEdge[index],0.0,1.0,0.0);
		}
		const U32 primitiveCount=spOutputPrimitives->count();
		for(U32 index=0;index<primitiveCount;index++)
		{
			const U32 subCount=spOutputElements->subCount(index);
			if(subCount<2)
			{
				continue;
			}
			SpatialVector lastPoint=
					spOutputPrimitives->spatialVector(index,subCount-1);
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const I32 vertexIndex=
						spOutputPrimitives->integer(index,subIndex);
				const SpatialVector point=
						spOutputPrimitives->spatialVector(index,subIndex);
				pOneEdge[vertexIndex]=unit(point-lastPoint);
			}
		}
	}

	for(U32 index=0;index<elementCount;index++)
	{
		const U32 subCount=spOutputElements->subCount(index);
		if(subCount<1)
		{
			continue;
		}

		const SpatialVector point=spOutputElements->spatialVector(index,0);
		if(spiderWrap)
		{
			const SpatialVector norm=spOutputNormals->spatialVector(index,0);
			SpatialVector tangent=pOneEdge[index];
			SpatialVector side;
			cross(side,norm,tangent);
			cross(tangent,side,norm);

			SpatialVector sample[4];
			sample[0]=point;
			sample[1]=point+tangent;
			sample[2]=point-tangent+side;
			sample[3]=point-tangent-side;

			feLog("oneEdge %s tangent %s\n",
					c_print(pOneEdge[index]),c_print(tangent));

			for(U32 binding=0;binding<4;binding++)
			{
				std::vector< sp<SurfaceI::ImpactI> > impactArray=
						spDriver->nearestPoints(sample[binding],1);
				if(impactArray.size())
				{
					sp<SurfaceI::ImpactI>& rspImpact=impactArray[0];

					face=rspImpact->face();
					barycenter=rspImpact->barycenter();
				}
				else
				{
					face= -1;
					set(barycenter);
				}

				outputFaces[binding]->set(index,face);
				outputBarys[binding]->set(index,barycenter);
			}
		}
		else
		{
			std::vector< sp<SurfaceI::ImpactI> > impactArray=
					spDriver->nearestPoints(point,bindings);
			const U32 impacts=impactArray.size();

			for(U32 binding=0;binding<bindings;binding++)
			{
				if(binding<impacts && impactArray[binding].isValid())
				{
					sp<SurfaceI::ImpactI>& rspImpact=impactArray[binding];

					face=rspImpact->face();
					barycenter=rspImpact->barycenter();
				}
				else
				{
					face= -1;
					set(barycenter);
				}

				outputFaces[binding]->set(index,face);
				outputBarys[binding]->set(index,barycenter);
			}
		}
	}

	delete[] pOneEdge;

//	feLog("SurfaceBindOp::handle done\n");
}

I32 SurfaceBindOp::accessBindings(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	std::vector< sp<SurfaceAccessorI> >&a_rFaces,
	std::vector< sp<SurfaceAccessorI> >&a_rBarys,
	BWORD a_allowCreation)
{
	const String bindPrefix=catalog<String>("Bind Attribute Prefix");

	const bool spiderWrap=catalog<bool>("Spider");
	const I32 bindings=spiderWrap? 4: catalog<I32>("Bindings");
	const bool bindPrimitives=catalog<bool>("Bind Primitives");
	const SurfaceAccessibleI::Element element=bindPrimitives?
			SurfaceAccessibleI::e_primitive:
			SurfaceAccessibleI::e_point;

	a_rFaces.resize(bindings);
	a_rBarys.resize(bindings);
	U32 bindCount=0;
	for(U32 binding=0;binding<bindings;binding++)
	{
		String faceName;
		String baryName;
		faceName.sPrintf("%sFace%d",bindPrefix.c_str(),binding);
		baryName.sPrintf("%sBary%d",bindPrefix.c_str(),binding);

		a_rFaces[binding]=a_rspAccessibleI->accessor(element,
				faceName,a_allowCreation);
		if(a_rFaces[binding].isNull())
		{
			if(a_allowCreation)
			{
				feLog("SurfaceBindOp::handle face attribute"
						" \"%s\" not accessible\n",faceName.c_str());
				catalog<String>("error")="inaccessible face attribute \""+
							faceName+"\"";
				return -1;
			}
			break;
		}

		a_rBarys[binding]=a_rspAccessibleI->accessor(element,
				baryName,a_allowCreation);
		if(a_allowCreation && a_rBarys[binding].isNull())
		{
			if(a_allowCreation)
			{
				feLog("SurfaceBindOp::handle barycentric attribute"
						" \"%s\" not accessible\n",baryName.c_str());
				catalog<String>("error")=
							"inaccessible barycentric attribute \""+
							baryName+"\"";
				return -1;
			}
			break;
		}

		bindCount++;
	}

	return bindCount;
}
