/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "vegetation/vegetation.pmh"

using namespace fe;

TreeOp::TreeOp(void)
{
}

void TreeOp::initialize(void)
{
	catalog<String>("Tree Name")="Aspen_Quaking";

	catalog<SpatialVector>("Location");

	catalog<SpatialVector>("Wind Direction");

	catalog<bool>("Y Is Up")=true;

	catalog<SpatialVector>("Center")=SpatialVector(0.0,0.0,0.0);

	catalog<Real>("Radius")=1e0;
	catalog<Real>("Radius","high")=1e1;
}

TreeOp::~TreeOp(void)
{
#if FE_CODEGEN<=FE_DEBUG
	Peeker peeker;
	if(m_spSignalerI.isValid())
	{
		m_spSignalerI->peek(peeker);
	}

	feLog("\n~TreeOp\n%s\n",peeker.str().c_str());
#endif
}

void TreeOp::bind(sp<SignalerI> spSignalerI,sp<Layout> spLayout)
{
	feLog("TreeOp::bind\n");
	FEASSERT(spSignalerI.isValid());
	FEASSERT(spLayout.isValid());

	OperatorSurfaceCommon::bind(spSignalerI,spLayout);

	setName(catalog<String>("Tree Name"));

	sp<Registry> spRegistry=spSignalerI->registry();

	m_spSignalerI=spRegistry->create("SignalerI");
	sp<HandlerI> spSimPlant=spRegistry->create("*.SimPlant");
	sp<HandlerI> spDrawPlant=spRegistry->create("*.DrawPlant");
	if(!m_spSignalerI.isValid() || !spSimPlant.isValid() ||
			!spDrawPlant.isValid())
	{
		feX("TreeOp::initialize", "can't create components");
	}
	m_modifierArray.push_back(spSimPlant);
	m_modifierArray.push_back(spDrawPlant);

	sp<Scope> spScope=spRegistry->master()->catalog()->catalogComponent(
			"Scope","PlantScope");
	FEASSERT(spScope.isValid());

	m_forestRV.bind(spScope);
	m_forestRV.createRecord();
	set(m_forestRV.effectorVelocity());
	set(m_forestRV.windSource());

	m_plantRV.bind(spScope);

	m_spSignalerI->insert(spSimPlant,m_forestRV.layout());
	m_spSignalerI->insert(spDrawPlant,m_forestRV.layout());

	String longname;
	longname.sPrintf("../media/vegetation/%s.rg",name().c_str());
	sp<RecordGroup>	spRG=RecordView::loadRecordGroup(spScope,longname);
	RecordGroup::iterator it=spRG->begin();
	sp<RecordArray>& rspRA= *it;
	if(m_plantRV.plantName.check(rspRA))
	{
		m_forestRV.plants()->add(spRG);
	}
	else
	{
		Record plant=m_plantRV.createRecord();
		m_plantRV.plantName()=name();
		m_plantRV.location()=catalog<SpatialVector>("Location");
		m_forestRV.plants()->add(plant);
	}

	m_plantRV.bind(NULL);
}

void TreeOp::handle(Record& a_rSignal)
{
	feLog("TreeOp::handle\n");

	sp<DrawI> spDrawI=drawI(a_rSignal);
	if(!spDrawI.isValid())
	{
		feLog("TreeOp::handle DrawI invalid\n");
		catalog<String>("error")="invalid drawing interface";
		return;
	}

	bool yIsUp=catalog<bool>("Y Is Up");

	m_forestRV.upAxis()=yIsUp? 1: 2;
	m_forestRV.drawI()=spDrawI;
	m_forestRV.deltaT()=(1.0f/24.0f);
	m_forestRV.windDirection()=catalog<SpatialVector>("Wind Direction");

	const SpatialVector center=catalog<SpatialVector>("Center");
	const Real radius=catalog<Real>("Radius");
	Vector4 effector;
	if(yIsUp)
	{
		set(effector,center[0],-center[2],center[1],radius);
	}
	else
	{
		effector=center;
		effector[3]=radius;
	}
	m_forestRV.effector()=effector;

	sp<RecordGroup>& rspPlants=m_forestRV.plants();
	FEASSERT(rspPlants.isValid());

	//* update location, but only if it is an individual tree
	for(RecordGroup::iterator it=rspPlants->begin();it!=rspPlants->end();it++)
	{
		sp<RecordArray> spRA= *it;

		if(spRA->length()!=1)
		{
			continue;
		}

		if(!m_plantRV.plantModel.check(spRA))
		{
			continue;
		}

		m_plantRV.bind(spRA);
		while(m_plantRV.step())
		{
			m_plantRV.location()=catalog<SpatialVector>("Location");
		}
	}

	m_plantRV.bind(NULL);

	Record forest=m_forestRV.record();
	m_spSignalerI->signal(forest);

	feLog("TreeOp::handle done\n");
}
